@@ -27,6 +27,9 @@ char *JOIN_CT_ENV[] =  { "OMPI_COMM_WORLD_LOCAL_SIZE",
 char *JOIN_TAG_ENV[] = { "SLURM_STEP_ID",
                          NULL };
 
+/* Default overlaid tmpfs size. */
+char *WRITE_FAKE_DEFAULT = "12%";
+
 
 /** Command line options **/
 
@@ -41,7 +44,7 @@ Example:\n\
 \n\
 You cannot use this program to actually change your UID.\n";
 
-const char args_doc[] = "IMAGE -- CMD [ARG...]";
+const char args_doc[] = "IMAGE -- COMMAND [ARG...]";
 
 /* Note: Long option numbers, once issued, are permanent; i.e., if you remove
    one, don’t re-number the others. */
@@ -57,9 +60,11 @@ const struct argp_option options[] = {
    { "join-pid",       -5, "PID",  0, "join a namespace using a PID" },
    { "join-ct",        -3, "N",    0, "number of join peers (implies --join)" },
    { "join-tag",       -4, "TAG",  0, "label for peer group (implies --join)" },
+   { "test",          -17, "TEST", 0, "do test TEST" },
    { "mount",         'm', "DIR",  0, "SquashFS mount point"},
    { "no-passwd",      -9, 0,      0, "don't bind-mount /etc/{passwd,group}"},
    { "private-tmp",   't', 0,      0, "use container-private /tmp" },
+   { "quiet",         'q', 0,      0, "print less output (can be repeated)"},
 #ifdef HAVE_SECCOMP
    { "seccomp",       -14, 0,      0,
                            "fake success for some syscalls with seccomp(2)"},
@@ -75,7 +80,9 @@ const struct argp_option options[] = {
    { "verbose",       'v', 0,      0, "be more verbose (can be repeated)" },
    { "version",       'V', 0,      0, "print version and exit" },
    { "warnings",      -16, "NUM",  0, "log NUM warnings and exit" },
-   { "write",         'w', 0,      0, "mount image read-write"},
+   { "write",         'w', 0,      0, "mount image read-write (avoid)"},
+   { "write-fake",    'W', "SIZE", OPTION_ARG_OPTIONAL,
+                           "overlay read-write tmpfs on top of image" },
    { 0 }
 };
 
@@ -155,6 +162,7 @@ int main(int argc, char *argv[])
                                .join_ct = 0,
                                .join_pid = 0,
                                .join_tag = NULL,
+                               .overlay_size = NULL,
                                .private_passwd = false,
                                .private_tmp = false,
                                .type = IMG_NONE,
@@ -173,13 +181,16 @@ int main(int argc, char *argv[])
       argp_help_fmt_set = true;
    else {
       argp_help_fmt_set = false;
-      Z_ (setenv("ARGP_HELP_FMT", "opt-doc-col=25,no-dup-args-note", 0));
+      Z_ (setenv("ARGP_HELP_FMT", "opt-doc-col=27,no-dup-args-note", 0));
    }
    Z_ (argp_parse(&argp, argc, argv, 0, &arg_next, &args));
    if (!argp_help_fmt_set)
       Z_ (unsetenv("ARGP_HELP_FMT"));
 
-   Te (arg_next < argc - 1, "IMAGE and/or CMD not specified");
+   if (arg_next >= argc - 1) {
+      printf("usage: ch-run [OPTION...] IMAGE -- COMMAND [ARG...]\n");
+      FATAL("IMAGE and/or COMMAND not specified");
+   }
    args.c.img_ref = argv[arg_next++];
    args.c.newroot = realpath_(args.c.newroot, true);
    args.storage_dir = realpath_(args.storage_dir, true);
@@ -454,6 +465,10 @@ static error_t parse_opt(int key, char *arg, struct argp_state *state)
       break;
    case -12: // --home
       Tf (args->c.host_home = getenv("HOME"), "--home failed: $HOME not set");
+      if (args->c.overlay_size == NULL) {
+         VERBOSE("--home specified; also setting --write-fake");
+         args->c.overlay_size = WRITE_FAKE_DEFAULT;
+      }
       break;
    case -13: // --unsafe
       args->unsafe = true;
@@ -463,13 +478,21 @@ static error_t parse_opt(int key, char *arg, struct argp_state *state)
       args->seccomp_p = true;
       break;
 #endif
+   case -15: // --set-env0
+      parse_set_env(args, arg, '\0');
+      break;
    case -16: // --warnings
       for (int i = 1; i <= parse_int(arg, false, "--warnings"); i++)
          WARNING("this is warning %d!", i);
       exit(0);
       break;
-   case -15: // --set-env0
-      parse_set_env(args, arg, '\0');
+   case -17: // --test
+      if (!strcmp(arg, "log"))
+         test_logging(false);
+      else if (!strcmp(arg, "log-fail"))
+         test_logging(true);
+      else
+         FATAL("invalid --test argument: %s; see source code", arg);
       break;
    case 'b': {  // --bind
          char *src, *dst;
@@ -512,6 +535,11 @@ static error_t parse_opt(int key, char *arg, struct argp_state *state)
       if (!path_exists(arg, NULL, false))
          WARNING("storage directory not found: %s", arg);
       break;
+   case 'q':  // --quiet
+      Te(verbose <= 0, "--quiet incompatible with --verbose");
+      verbose--;
+      Te(verbose >= -3, "--quiet can be specified at most thrice");
+      break;
    case 't':  // --private-tmp
       args->c.private_tmp = true;
       break;
@@ -525,12 +553,16 @@ static error_t parse_opt(int key, char *arg, struct argp_state *state)
       exit(EXIT_SUCCESS);
       break;
    case 'v':  // --verbose
+      Te(verbose >= 0, "--verbose incompatible with --quiet");
       verbose++;
       Te(verbose <= 3, "--verbose can be specified at most thrice");
       break;
    case 'w':  // --write
       args->c.writable = true;
       break;
+   case 'W':  // --write-fake
+      args->c.overlay_size = arg != NULL ? arg : WRITE_FAKE_DEFAULT;
+      break;
    case ARGP_KEY_NO_ARGS:
       argp_state_help(state, stderr, (  ARGP_HELP_SHORT_USAGE
                                       | ARGP_HELP_PRE_DOC
