@@ -1,4 +1,5 @@
 import getpass
+import hashlib
 import io
 import os
 import re
@@ -48,40 +49,6 @@ auth_p = False
 
 ## Classes ##
 
-class Credentials:
-
-   __slots__ = ("password",
-                "username")
-
-   def __init__(self):
-      self.username = None
-      self.password = None
-
-   def get(self):
-      # If stored, return those.
-      if (self.username is not None):
-         username = self.username
-         password = self.password
-      else:
-         try:
-            # Otherwise, use environment variables.
-            username = os.environ["CH_IMAGE_USERNAME"]
-            password = os.environ["CH_IMAGE_PASSWORD"]
-         except KeyError:
-            # Finally, prompt the user.
-            # FIXME: This hangs in Bats despite sys.stdin.isatty() == True.
-            try:
-               username = input("\nUsername: ")
-            except KeyboardInterrupt:
-               ch.FATAL("authentication cancelled")
-            password = getpass.getpass("Password: ")
-         if (not ch.password_many):
-            # Remember the credentials.
-            self.username = username
-            self.password = password
-      return (username, password)
-
-
 class Auth(requests.auth.AuthBase):
 
    # Every registry request has an “authorization object”. This starts as no
@@ -111,13 +78,6 @@ class Auth(requests.auth.AuthBase):
 
    __slots__ = ("auth_h_next",)  # WWW-Authenticate header for next escalator
 
-   def __eq__(self, other):
-      return (type(self) == type(other))
-
-   @property
-   def escalators(self):
-      ...
-
    @classmethod
    def authenticate(class_, creds, auth_d):
       """Authenticate using the given credentials and parsed WWW-Authenticate
@@ -125,6 +85,13 @@ class Auth(requests.auth.AuthBase):
          not. The caller is responsible for dealing with the failure."""
       ...
 
+   def __eq__(self, other):
+      return (type(self) == type(other))
+
+   @property
+   def escalators(self):
+      ...
+
    def escalate(self, reg, res):
       """Escalate to a higher level of authorization. Use the WWW-Authenticate
          header in failed response res if there is one."""
@@ -172,6 +139,16 @@ class Auth_Basic(Auth):
 
    __slots__ = ("basic")
 
+   @classmethod
+   def authenticate(class_, reg, auth_d):
+      # Note: Basic does not validate the credentials until we try to use it.
+      if ("realm" not in auth_d):
+         ch.FATAL("WWW-Authenticate missing realm")
+      (username, password) = reg.creds.get()
+      i = class_()
+      i.basic = requests.auth.HTTPBasicAuth(username, password)
+      return i
+
    def __call__(self, *args, **kwargs):
       return self.basic(*args, **kwargs)
 
@@ -185,16 +162,6 @@ class Auth_Basic(Auth):
    def escalators(self):
       return ()
 
-   @classmethod
-   def authenticate(class_, reg, auth_d):
-      # Note: Basic does not validate the credentials until we try to use it.
-      if ("realm" not in auth_d):
-         ch.FATAL("WWW-Authenticate missing realm")
-      (username, password) = reg.creds.get()
-      i = class_()
-      i.basic = requests.auth.HTTPBasicAuth(username, password)
-      return i
-
 
 class Auth_Bearer_IDed(Auth):
    # https://stackoverflow.com/a/58055668
@@ -210,28 +177,6 @@ class Auth_Bearer_IDed(Auth):
       self.token = token
       self.auth_d = auth_d
 
-   def __call__(self, req):
-      req.headers["Authorization"] = "Bearer %s" % self.token
-      return req
-
-   def __eq__(self, other):
-      return super().__eq__(other) and (self.auth_d == other.auth_d)
-
-   def __str__(self):
-      return ("Bearer (%s) %s" % (self.__class__.__name__.split("_")[-1],
-                                  self.token_short))
-
-   @property
-   def escalators(self):
-      # One can escalate to an authenticated Bearer with a greater scope. I’m
-      # pretty sure this doesn’t create an infinite loop because eventually
-      # the token request will fail.
-      return (Auth_Bearer_IDed,)
-
-   @property
-   def token_short(self):
-      return ("%s..%s" % (self.token[:8], self.token[-8:]))
-
    @classmethod
    def authenticate(class_, reg, auth_d):
       # Registries and endpoints vary in what they put in WWW-Authenticate. We
@@ -259,6 +204,28 @@ class Auth_Bearer_IDed(Auth):
       (username, password) = creds.get()
       return requests.auth.HTTPBasicAuth(username, password)
 
+   def __call__(self, req):
+      req.headers["Authorization"] = "Bearer %s" % self.token
+      return req
+
+   def __eq__(self, other):
+      return super().__eq__(other) and (self.auth_d == other.auth_d)
+
+   def __str__(self):
+      return ("Bearer (%s) %s" % (self.__class__.__name__.split("_")[-1],
+                                  self.token_short))
+
+   @property
+   def escalators(self):
+      # One can escalate to an authenticated Bearer with a greater scope. I’m
+      # pretty sure this doesn’t create an infinite loop because eventually
+      # the token request will fail.
+      return (Auth_Bearer_IDed,)
+
+   @property
+   def token_short(self):
+      return ("%s..%s" % (self.token[:8], self.token[-8:]))
+
 
 class Auth_Bearer_Anon(Auth_Bearer_IDed):
    anon_p = True
@@ -267,16 +234,16 @@ class Auth_Bearer_Anon(Auth_Bearer_IDed):
 
    __slots__ = ()
 
-   @property
-   def escalators(self):
-      return (Auth_Bearer_IDed,)
-
    @classmethod
    def token_auth(class_, creds):
       # The way to get an anonymous Bearer token is to give no Basic auth
       # header in the token request.
       return None
 
+   @property
+   def escalators(self):
+      return (Auth_Bearer_IDed,)
+
 
 class Auth_None(Auth):
    anon_p = True
@@ -296,6 +263,40 @@ class Auth_None(Auth):
               Auth_Bearer_IDed)
 
 
+class Credentials:
+
+   __slots__ = ("password",
+                "username")
+
+   def __init__(self):
+      self.username = None
+      self.password = None
+
+   def get(self):
+      # If stored, return those.
+      if (self.username is not None):
+         username = self.username
+         password = self.password
+      else:
+         try:
+            # Otherwise, use environment variables.
+            username = os.environ["CH_IMAGE_USERNAME"]
+            password = os.environ["CH_IMAGE_PASSWORD"]
+         except KeyError:
+            # Finally, prompt the user.
+            # FIXME: This hangs in Bats despite sys.stdin.isatty() == True.
+            try:
+               username = input("\nUsername: ")
+            except KeyboardInterrupt:
+               ch.FATAL("authentication cancelled")
+            password = getpass.getpass("Password: ")
+         if (not ch.password_many):
+            # Remember the credentials.
+            self.username = username
+            self.password = password
+      return (username, password)
+
+
 class HTTP:
    """Transfers image data to and from a remote image repository via HTTPS.
 
@@ -395,7 +396,7 @@ class HTTP:
       # /v2/library/hello-world/blobs/<layer-hash>
       url = self._url_of("blobs", "sha256:" + digest)
       sw = ch.Progress_Writer(path, msg)
-      self.request("GET", url, out=sw)
+      self.request("GET", url, out=sw, hd=digest)
       sw.close()
 
    def blob_upload(self, digest, data, note=""):
@@ -488,7 +489,7 @@ class HTTP:
       "Upload gzipped tarball layer at path, which must have hash digest."
       # NOTE: We don’t verify the digest b/c that means reading the whole file.
       ch.VERBOSE("layer tarball: %s" % path)
-      fp = path.open_("rb") # open file avoids reading it all into memory
+      fp = path.open("rb")  # open file avoids reading it all into memory
       self.blob_upload(digest, fp, note)
       ch.close_(fp)
 
@@ -518,17 +519,19 @@ class HTTP:
       self.request("PUT", url, {201}, data=manifest,
                    headers={ "Content-Type": TYPES_MANIFEST["docker2"] })
 
-   def request(self, method, url, statuses={200}, out=None, **kwargs):
+   def request(self, method, url, statuses={200}, out=None, hd=None, **kwargs):
       """Request url using method and return the response object. If statuses
          is given, it is set of acceptable response status codes, defaulting
          to {200}; any other response is a fatal error. If out is given,
          response content will be streamed to this Progress_Writer object and
-         must be non-zero length.
+         must be non-zero length. If hd is given, validate integrity of
+         downloaded data using expected hash digest.
 
          Use current session if there is one, or start a new one if not. If
          authentication fails (or isn’t initialized), then authenticate harder
          and re-try the request."""
       # Set up.
+      assert (out or hd is None), "digest only checked if streaming"
       self.session_init_maybe()
       ch.VERBOSE("auth: %s" % self.auth)
       if (out is not None):
@@ -547,6 +550,7 @@ class HTTP:
             else:
                ch.FATAL("unhandled authentication failure")
       # Stream response if needed.
+      m = hashlib.sha256()
       if (out is not None and res.status_code == 200):
          try:
             length = int(res.headers["Content-Length"])
@@ -557,6 +561,10 @@ class HTTP:
          out.start(length)
          for chunk in res.iter_content(ch.HTTP_CHUNK_SIZE):
             out.write(chunk)
+            m.update(chunk) # store downloaded hash digest
+         # Validate integrity of downloaded data
+         if (hd is not None and hd != m.hexdigest()):
+            ch.FATAL("registry streamed response content is invalid")
       # Done.
       return res
 
